import new
import copy

class InvokeMacro:
    def __init__(self, name, args):
        self.name = name
        self.args = args
        
    def _find(self, n, par_args, globals, par_envs):
        if n in par_args:
            return par_args[n]
        else:
            # Create the prefixes
            prefixes = [""]
            prefix = ""
            envs = par_envs[0:]
            
            while len(envs) > 0:
                prefix = "%s%s_" % (prefix, envs.pop(0))
                prefixes.insert(0, prefix)
        
            for prefix in prefixes:
                name = "%s%s" % (prefix, n)
                if name in globals:
                    return globals[name]
            
            print "[WARN] Cannot find name: %s, prefixes: %r" % (name, prefixes)
            return name
        
    def invoke(self, level, par_args, globals, par_envs):
    
        # Find the macro
        #if self.name in self.args:
        #    return self.args[self.name].run(level, self.args, globals, par_envs)
        
        macro = self._find(self.name, par_args, globals, par_envs)
        if isinstance(macro, Macro):
            args = {}
            for i in range(0, len(macro.arg_names)):
                arg = None
                if i < len(self.args):
                    if isinstance(self.args[i], str):
                        arg = self._find(self.args[i], par_args, globals, par_envs)
                    else:
                        arg = self.args[i]
                
                args[macro.arg_names[i]] = arg
                    
            globs = copy.deepcopy(globals)
            for key in par_args:
                globs[key] = par_args[key]
            return macro.run(level, args, globs, par_envs)
        else:
            return macro
        
    def __repr__(self):
        out = self.name
        for arg in self.args:
            out = "%s,%r" % (out, arg)
        return "<%s>" % out
        
class Macro:
    def __init__(self, body):
        self.body = body.strip()
        self.arg_names = []
        self.body_code = None
        self.name = None

    def createCode(self):
        if self.body_code == None:
            self.body_code = ""
            lines = self.body.split("\n")
            lines = map(lambda a: "    " + a, lines)
            self.body_code = "def _f(self):\n" + "\n".join(lines)
        
    def run(self, level, args, globals, envs):
        self.createCode()
        
        #print "Running: %s(%s)" % (self.name, self.__class__)
        call_globals = {}
        for key in globals:
            call_globals[key] = globals[key]
        for name in self.arg_names:
            call_globals[name] = args[name]
            
        call_globals["_level"]      = level
        call_globals["_args"]       = args
        call_globals["_globals"]    = globals
        call_globals["_envs"]       = envs
        call_globals["Macro"]       = Macro
        
        #print "Globals: %s" % call_globals
        #print "Args: %s"    % args
        
        exec self.body_code in call_globals
        rv = call_globals['_f'](self)
        #print "    COMPLETE"
        return rv
        
    def __repr__(self):
        return '<%r %r>' % (self.name, self.arg_names)

class MacroNumber(Macro):
    def __init__(self, number):
        body = '''return ('%g' % self.value)'''
        Macro.__init__(self, body)
        self.value = number
        self.name = '%g' % number
        
    def __repr__(self):
        return '%g' % self.value

class MacroPair(Macro):
    def __init__(self, init, incr):
        body = '''return str(self.init + self.incr * (_level - 1))'''
        Macro.__init__(self, body)
        self.init = init
        self.incr = incr
        
    def __repr__(self):
        return '%g(%g)' % (self.init, self.incr)

class MacroArray(Macro):
    def __init__(self, contents):
        body = '''
lvl = min(len(self.contents), _level)
if lvl > 0:
    val = self.contents[lvl - 1]
    if isinstance(val, Macro):
        val = val.run(_level, _args, _globals, _envs)
    elif not isinstance(val, str):
        val = str(val)
    return val
else:
    return ""
'''
        Macro.__init__(self, body)
        self.contents = contents
        
    def __repr__(self):
        return repr(self.contents)
        
class MacroText(Macro):
    def __init__(self):
        body = '''
out = ''
for component in self.components:
    if isinstance(component, str):
        str_comp = component
    else:
        str_comp = component.invoke(_level, _args, _globals, _envs)
    out = out + str_comp
return out
'''
        Macro.__init__(self, body)
        self.components = []
        
    def add(self, component):
        self.components.insert(0, component)
        
    def __repr__(self):
        return repr(self.components)
        
class Program:
    def __init__(self, init_macros):
        self.macros = init_macros
        self.texts = {}
        
    def add(self, other):
        self.addMacros(other.macros)
        self.addTexts(other.texts)

    def addMacros(self, macros):
        for key in macros:
            self.macros[key] = macros[key]
            
    def addTexts(self, texts):
        for key in texts:
            self.texts[key] = texts[key]
            
    def find(self, code, level):
        envs = code.split("_")
        if code in self.texts:
            return self.texts[code].run(level, {}, self.macros, envs)
        else:
            return code
            
    def __repr__(self):
        out = "macros:"
        keys = self.macros.keys()
        keys.sort()
        for key in keys:
            out = '''%s
    %r\t%r''' % (out, key, self.macros[key])
        out = '%s\n\ntexts:' % out
        
        keys = self.texts.keys()
        keys.sort()
        for key in keys:
            out = '''%s
    %r\t%r''' % (out, key, self.texts[key])
        return out